@@ -1,12 +1,16 @@
-use std::ffi;
+#[cfg(feature = "v1_18")]
+use std::num::NonZeroU16;
 use std::os::raw::c_void;
-use std::ptr;
+use std::{ffi, ptr};
 
 use four_cc::FourCC;
 use libheif_sys as lh;
 
+use crate::encoder::get_encoding_options_ptr;
 use crate::reader::{Reader, HEIF_READER};
 use crate::utils::str_to_cstring;
+#[cfg(feature = "v1_19")]
+use crate::SecurityLimits;
 use crate::{
     Encoder, EncodingOptions, HeifError, HeifErrorCode, HeifErrorSubCode, Image, ImageHandle,
     ItemId, Result,
@@ -46,25 +50,48 @@ impl HeifContext<'static> {
     /// Create a new context from file.
     pub fn read_from_file(name: &str) -> Result<HeifContext<'static>> {
         let mut context = HeifContext::new()?;
-        context.source = Source::File;
+        context.read_file(name)?;
+        Ok(context)
+    }
+
+    /// Read a HEIF file from a named disk file.
+    pub fn read_file(&mut self, name: &str) -> Result<()> {
+        self.source = Source::File;
         let c_name = ffi::CString::new(name).unwrap();
         let err =
-            unsafe { lh::heif_context_read_from_file(context.inner, c_name.as_ptr(), ptr::null()) };
+            unsafe { lh::heif_context_read_from_file(self.inner, c_name.as_ptr(), ptr::null()) };
         HeifError::from_heif_error(err)?;
-        Ok(context)
+        Ok(())
     }
 
     /// Create a new context from reader.
     pub fn read_from_reader(reader: Box<dyn Reader>) -> Result<HeifContext<'static>> {
         let mut context = HeifContext::new()?;
+        context.read_reader(reader)?;
+        Ok(context)
+    }
+
+    /// Read a HEIF file from the reader.
+    pub fn read_reader(&mut self, reader: Box<dyn Reader>) -> Result<()> {
         let mut reader_box = Box::new(reader);
         let user_data = reader_box.as_mut() as *mut _ as *mut c_void;
         let err = unsafe {
-            lh::heif_context_read_from_reader(context.inner, &HEIF_READER, user_data, ptr::null())
+            lh::heif_context_read_from_reader(self.inner, &HEIF_READER, user_data, ptr::null())
         };
         HeifError::from_heif_error(err)?;
-        context.source = Source::Reader(reader_box);
-        Ok(context)
+        self.source = Source::Reader(reader_box);
+        Ok(())
+    }
+
+    /// # Safety
+    ///
+    /// The given pointer must be valid.
+    #[cfg(feature = "v1_18")]
+    pub(crate) unsafe fn from_ptr(ctx: *mut lh::heif_context) -> HeifContext<'static> {
+        HeifContext {
+            inner: ctx,
+            source: Source::None,
+        }
     }
 }
 
@@ -74,19 +101,29 @@ impl<'a> HeifContext<'a> {
     /// The provided memory buffer is not copied.
     /// That means, you will have to keep the memory buffer alive as
     /// long as you use the context.
-    pub fn read_from_bytes(bytes: &[u8]) -> Result<HeifContext> {
+    pub fn read_from_bytes(bytes: &[u8]) -> Result<HeifContext<'_>> {
         let mut context = HeifContext::new()?;
-        context.source = Source::Memory(bytes);
+        context.read_bytes(bytes)?;
+        Ok(context)
+    }
+
+    /// Read a HEIF file from bytes.
+    ///
+    /// The provided memory buffer is not copied.
+    /// That means you will have to keep the memory buffer alive as
+    /// long as you use the context.
+    pub fn read_bytes<'b: 'a>(&mut self, bytes: &'b [u8]) -> Result<()> {
+        self.source = Source::Memory(bytes);
         let err = unsafe {
             lh::heif_context_read_from_memory_without_copy(
-                context.inner,
+                self.inner,
                 bytes.as_ptr() as _,
                 bytes.len(),
                 ptr::null(),
             )
         };
         HeifError::from_heif_error(err)?;
-        Ok(context)
+        Ok(())
     }
 
     unsafe extern "C" fn vector_writer(
@@ -188,17 +225,13 @@ impl<'a> HeifContext<'a> {
         encoder: &mut Encoder,
         encoding_options: Option<EncodingOptions>,
     ) -> Result<ImageHandle> {
-        let encoding_options_ptr = match encoding_options {
-            Some(options) => options.inner,
-            None => ptr::null(),
-        };
         let mut handle: *mut lh::heif_image_handle = ptr::null_mut();
         unsafe {
             let err = lh::heif_context_encode_image(
                 self.inner,
                 image.inner,
                 encoder.inner,
-                encoding_options_ptr,
+                get_encoding_options_ptr(&encoding_options),
                 &mut handle,
             );
             HeifError::from_heif_error(err)?;
@@ -223,10 +256,6 @@ impl<'a> HeifContext<'a> {
         encoder: &mut Encoder,
         encoding_options: Option<EncodingOptions>,
     ) -> Result<Option<ImageHandle>> {
-        let encoding_options_ptr = match encoding_options {
-            Some(options) => options.inner,
-            None => ptr::null(),
-        };
         let mut handle: *mut lh::heif_image_handle = ptr::null_mut();
         unsafe {
             let err = lh::heif_context_encode_thumbnail(
@@ -234,7 +263,7 @@ impl<'a> HeifContext<'a> {
                 image.inner,
                 master_image_handle.inner,
                 encoder.inner,
-                encoding_options_ptr,
+                get_encoding_options_ptr(&encoding_options),
                 bbox_size.min(i32::MAX as _) as _,
                 &mut handle,
             );
@@ -243,6 +272,46 @@ impl<'a> HeifContext<'a> {
         Ok(Some(ImageHandle::new(handle)))
     }
 
+    /// Encodes an array of images into a grid.
+    ///
+    /// # Arguments
+    ///
+    /// * `tiles` - User allocated array of images that will form the grid.
+    /// * `rows` - The number of rows in the grid. The number of columns will
+    ///   be calculated from the size of `tiles`.
+    /// * `encoder` - Defines the encoder to use.
+    ///   See [LibHeif::encoder_for_format()](crate::LibHeif::encoder_for_format).
+    /// * `encoding_options` - Optional, may be None.
+    ///
+    /// Returns an error if `tiles` slice is empty.
+    #[cfg(feature = "v1_18")]
+    pub fn encode_grid(
+        &mut self,
+        tiles: &[Image],
+        rows: NonZeroU16,
+        encoder: &mut Encoder,
+        encoding_options: Option<EncodingOptions>,
+    ) -> Result<Option<ImageHandle>> {
+        let mut handle: *mut lh::heif_image_handle = ptr::null_mut();
+        let mut tiles_inners: Vec<*mut lh::heif_image> =
+            tiles.iter().map(|img| img.inner).collect();
+        let rows = rows.get();
+        let columns = (tiles_inners.len() as u32 / rows as u32).min(u16::MAX as _) as u16;
+        unsafe {
+            let err = lh::heif_context_encode_grid(
+                self.inner,
+                tiles_inners.as_mut_ptr(),
+                rows,
+                columns,
+                encoder.inner,
+                get_encoding_options_ptr(&encoding_options),
+                &mut handle,
+            );
+            HeifError::from_heif_error(err)?;
+        }
+        Ok(Some(ImageHandle::new(handle)))
+    }
+
     /// Assign `master_image_handle` as the thumbnail image of `thumbnail_image_handle`.
     pub fn assign_thumbnail(
         &mut self,
@@ -332,7 +401,7 @@ impl<'a> HeifContext<'a> {
     /// decoded in the main thread. This is different from setting it to 1,
     /// which will generate a single background thread to decode the tiles.
     ///
-    /// Note that this setting only affects libheif itself. The codecs itself
+    /// Note that this setting only affects `libheif` itself. The codecs itself
     /// may still use multi-threaded decoding. You can use it, for example,
     /// in cases where you are decoding several images in parallel anyway you
     /// thus want to minimize parallelism in each decoder.
@@ -340,12 +409,30 @@ impl<'a> HeifContext<'a> {
         let max_threads = max_threads.min(libc::c_int::MAX as u32) as libc::c_int;
         unsafe { lh::heif_context_set_max_decoding_threads(self.inner, max_threads) };
     }
+
+    /// Returns the security limits for a context.
+    ///
+    /// By default, the limits are set to the global limits,
+    /// but you can change them with the help of [`HeifContext::set_security_limits()`] method.
+    #[cfg(feature = "v1_19")]
+    pub fn security_limits(&self) -> SecurityLimits {
+        let inner_ptr = unsafe { lh::heif_context_get_security_limits(self.inner) };
+        let inner = ptr::NonNull::new(inner_ptr).unwrap();
+        SecurityLimits::from_inner(inner)
+    }
+
+    /// Overwrites the security limits of a context.
+    #[cfg(feature = "v1_19")]
+    pub fn set_security_limits(&mut self, limits: &SecurityLimits) -> Result<()> {
+        let err = unsafe { lh::heif_context_set_security_limits(self.inner, limits.as_inner()) };
+        HeifError::from_heif_error(err)
+    }
 }
 
-impl<'a> Drop for HeifContext<'a> {
+impl Drop for HeifContext<'_> {
     fn drop(&mut self) {
         unsafe { lh::heif_context_free(self.inner) };
     }
 }
 
-unsafe impl<'a> Send for HeifContext<'a> {}
+unsafe impl Send for HeifContext<'_> {}
