@@ -1,3 +1,5 @@
+use std::collections::HashSet;
+
 use libheif_rs::LibHeif;
 
 #[test]
@@ -5,7 +7,15 @@ fn get_version() {
     let lib_heif = LibHeif::new();
     let version = lib_heif.version();
     assert!(version[0] >= 1);
-    assert!(version[1] >= 14);
+    if cfg!(feature = "v1_20") {
+        assert!(version[1] >= 20);
+    } else if cfg!(feature = "v1_19") {
+        assert!(version[1] >= 19);
+    } else if cfg!(feature = "v1_18") {
+        assert!(version[1] >= 18);
+    } else if cfg!(feature = "v1_17") {
+        assert!(version[1] >= 17);
+    }
 }
 
 #[test]
@@ -28,10 +38,18 @@ fn get_encoder() {
 fn get_encoder_for_format() {
     let lib_heif = LibHeif::new();
     let descriptors = lib_heif.encoder_descriptors(100, None, None);
+    let names: HashSet<String> = HashSet::from_iter(descriptors.iter().map(|d| d.name()));
     for descriptor in descriptors {
         let encoder = lib_heif
             .encoder_for_format(descriptor.compression_format())
             .unwrap();
-        assert_eq!(encoder.name(), descriptor.name());
+        assert!(names.contains(&encoder.name()));
     }
 }
+
+#[test]
+fn get_decoder_descriptors() {
+    let lib_heif = LibHeif::new();
+    let descriptors = lib_heif.decoder_descriptors(100, None);
+    assert!(descriptors.len() >= 2);
+}
